/*
** Module   : GenericWidget.java
** Abstract : common superclass for all server side widgets
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23492 Created initial version.
** 002 GES 20060114   @23910 Removed totally useless screen-value attr.
** 003 NVS 20060114   @24094 Implemented hide() and view() methods.
** 004 NVS 20060116   @24095 hide() and view() methods change visibility
**                           flag in widget config instance.
** 005 GES 20060203   @24261 Cache the widget ID in an instance var and
**                           require this to be set at the moment that
**                           the widget is added to the frame.  This is
**                           much faster than scanning the map (the old
**                           way).
** 006 GES 20060208   @24384 Added getFormat() and switched internal
**                           storage to be a string.
** 007 SIY 20060209   @24426 Minor cleanups and fixes.
** 008 NVS 20060310   @24964 Tracking ENTERED state using the ScreenBuffer
** 009 GES 20060315   @25080 Added missing setMultiple() signatures.
** 010 SIY 20060307   @25135 Changed signatures in get/setQuery.
** 011 SIY 20060307   @25136 Use ID declared in ComponentConfig.
** 012 NVS 20060321   @25178 Added isValidatable() method.
** 013 SIY 20060401   @25307 Added setHeader() method.
** 014 GES 20060413   @25515 Added getScreenValue/setScreenValue methods.
** 015 GES 20060426   @25733 Changed underline() implementation.
** 016 SIY 20060426   @25741 Changed signatures of get/setNextTabItem.
** 017 SIY 20060515   @26153 Moved COLUMN-LABEL attribute to 
**                           ControlEntity.
** 018 GES 20060803   @28379 Massive code cleanups to remove unneeded
**                           vars (less memory), unneeded processing
**                           (better performance), duplicated code and
**                           to meet coding standards. The only data
**                           left in this class is server-side data.
** 019 SIY 20060825   @28895 Mark frame as viewed if widget is enabled.
** 020 GES 20060925   @29869 Added removal of ampersand chars from label
**                           text to duplicate an undocumented 4GL
**                           feature.
** 021 GES 20061108   @31027 Added missing APIs for the list-item-pairs
**                           attribute.
** 022 NVS 20070426   @33244 Reimplemented SENSITIVE attribute support
**                           using new methods of LogicalTerminal.
** 023 GES 20080319   @37592 Added NEXT-SIBLING, PREV-SIBLING and TYPE.
** 024 GES 20080421   @38071 setFormat() doesn't call pushScreenDefinition
**                           but rather, it forces the format string
**                           change directly down to the client. This
**                           helps performance and allows the widget to
**                           redraw immediately which is needed to
**                           duplicate Progress behavior.
** 025 SIY 20080429   @38142 Added saveColors()/restoreColors() methods.
** 026 SIY 20080604   @38549 Added generation of error on any attempt to
**                           set too long label. Fixed comments.
** 027 SIY 20080609   @38649 Do not check label length until frame is 
**                           viewed.
** 028 SIY 20090210   @41287 Removed call to GenericFrame.setHadView(). Now 
**                           this method is invoked by LogicalTerminal.
** 029 CA  20090227   @41387 Added WrappedResources.isUnknown method; this
**                           implementation returns false.
** 030 CA  20090413   @41776 Fixed handling of VISIBLE/HIDDEN flag. When set
**                           to ON, it turns its counterpart to OFF and 
**                           vice versa.
** 031 GES 20090424   @41959 Import change.
** 032 SIY 20090701   @43023 Added getLabelStr() method.
** 033 SVL 20101004          Added support for DATA-TYPE, TABLE, TOOLTIP and
**                           WIDTH-PIXELS widget attributes.
** 034 SIY 20101004          Added more stubs.
** 035 SIY 20101005          Added stubs for HEIGHT-PIXELS getter/setter.
** 036 SVL 20101006          Added support for WIDTH, HEIGHT, TOOLTIP(String)
**                           attributes.
** 037 SVL 20101006          Added moveAfterTab, moveToTop, moveToBottom.
** 038 ECF 20101006          Changed signature of getName to return character
**                           rather than String.
** 039 SIY 20101014          Added asWidgetHandle() method.
** 040 SIY 20110112          Changes related to new color scheme.
** 041 SIY 20111101          Updated imports.
** 042 CA  20130112          getNextSibling and getPrevSibling must return a
**                           handle instance. Extracted into CommonWidget
**                           interface all widget APIs.
** 043 CS  20130204          Added conversion support for socket and 
**                           server-socket.Moved sesitive related methods from 
**                           CommonWidget in the Sensitive interface.
*/

package com.goldencode.p2j.ui;

import com.goldencode.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.util.*;

/**
 * This class represents server side widgets. Every widget is associated with
 * its configuration object, which is embedded into the widget. The
 * widget constructors create and initialize their backing configuration
 * objects.
 */
public class GenericWidget
implements WrappedResource,
           CommonWidget,
           Sensitive
{
   /** The containing frame object. */
   protected GenericFrame frame = null;
   
   /** The configuration container. */
   protected ComponentConfig config = null;

   /** The validator to use to validate proposed edits for this widget. */
   protected Validatable validator = null;

   /**
    * Constructor.
    *
    * @param    config
    *           The container for all state to be transferred to the client.
    */
   public GenericWidget(ComponentConfig config)
   {
      this.config = config;
   }

   /**
    * Sets the containing frame instance.
    * 
    * @param    hFrame
    *           handle which is expected to contain a <code>FrameWidget</code>
    *           instance
    */
   public void setFrame(handle hFrame)
   {
      // TODO:  no idea if this works...probably ClassCastException here
      setFrame((FrameWidget) hFrame.get());
   }

   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   public void setFrame(GenericFrame frame)
   {
      this.frame = frame;
   }

   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           <code>GenericFrame</code> instance that contains this widget
    */
   public void setFrame(FrameWidget frame)
   {
      this.frame = frame.getFrame();
   }
   
   /**
    * Gets the containing frame instance.
    * 
    * @return   The frame that contains this widget.
    */
   public GenericFrame getFrame()
   {
      return frame;
   }

   /**
    * Gets the numeric ID of this widget (this is used as an index into the
    * screen-buffer among other things).
    * 
    * @return   The numeric ID of this widget.
    */
   public int getId()
   {
      return config.getID();
   }

   /**
    * Sets the numeric ID of this widget (this is used as an index into the
    * screen-buffer among other things).
    * 
    * @param    id
    *           The numeric ID of this widget.
    */
   public void setId(int id)
   {
      config.setID(id);
   }

   /**
    * Checks whether the widget is changed during the most recent operation.
    * 
    * @return   <code>true</code> if this widget has changed its value.
    */
   public logical isEntered()
   {
      boolean entered = frame.getFrameBufferRef().isEntered(config.getID());
      return new logical(entered);
   }

   /**
    * Checks whether the widget is not changed during the most recent
    * operation.
    * 
    * @return   <code>true</code> if this widget has not changed its value.
    */
   public logical isNotEntered()
   {
      boolean entered = frame.getFrameBufferRef().isEntered(config.getID());
      return new logical(!entered);
   }

   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   public logical isReadOnly() 
   {
      return new logical(config.isReadOnly());
   }

   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   public void setReadOnly(logical r) 
   {
      config.setReadOnly(r.booleanValue());
      frame.pushScreenDefinition();
   }

   /**
    * Sets the AUTO-RETURN option.
    * 
    * @param    autoReturn
    *           <code>true</code> if the option is set to ON.
    */
   public void setAutoReturn(boolean autoReturn)
   {
      config.setAutoReturn(autoReturn);
      frame.pushScreenDefinition();
   }

   /**
    * Sets the BLANK option.
    * 
    * @param    blank
    *           <code>true</code> if the option is set to ON.
    */
   public void setBlank(boolean blank)
   {
      config.setBlank(blank);
      frame.pushScreenDefinition();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public decimal getColumn()
   {
      throw new RuntimeException("COLUMN attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   column
    *          Not used.
    */
   public void setColumn(NumberType column)
   {
      throw new RuntimeException("COLUMN attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   column
    *          Not used.
    */
   public void setColumn(double column)
   {
      throw new RuntimeException("COLUMN attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getDcolor()
   {
      throw new RuntimeException("DCOLOR attribute is undefined");
   }

   /**
    * Sets the display color.
    *
    * @param    param
    *           The display color.
    */
   public void setDcolor(NumberType param)
   {
      config.setDcolor(new ColorSpec(param.intValue()).convert());
      frame.pushScreenDefinition();

   }

   /**
    * Sets the display color.
    * 
    * @param    displayColor
    *           The display color.
    */
   public void setDcolor(Color displayColor)
   {
      config.setDcolor(displayColor);
      frame.pushScreenDefinition();
   }

   /**
    * Sets the DEBLANK option.
    * 
    * @param    deblank
    *           <code>true</code> if the option is set to ON.
    */
   public void setDeblank(boolean deblank)
   {
      // defined only for fill-ins
      ((FillInConfig) config).setDeblank(deblank);
      frame.pushScreenDefinition();
   }

   /**
    * Sets the DISABLE-AUTO-ZAP option.
    * 
    * @param    disableAutoZap
    *           <code>true</code> if the option is set to ON.
    */
   public void setDisableAutoZap(boolean disableAutoZap)
   {
      config.setDisableAutoZap(disableAutoZap);
      frame.pushScreenDefinition();
   }

   /**
    * Sets the format specification.
    * 
    * @param   format
    *          format specification
    */
   public void setFormat(character format)
   {
      setFormat(format.isUnknown() ? "" : format.getValue());
   }

   /**
    * Sets the format specification.
    * 
    * @param    format
    *           The format specification.
    */
   public void setFormat(String format)
   {
      setFormat(format, true);
   }

   /**
    * Gets the format specification.
    * 
    * @return   The format specification.
    */
   public String getFormat()
   {
      return config.getFormat();
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public decimal getHeightChars()
   {
      throw new RuntimeException("HEIGHT-CHARS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   h
    *          Not used.
    */
   public void setHeightChars(decimal h)
   {
      throw new RuntimeException("HEIGHT-CHARS attribute is undefined");
   }

   /**
    * Gets the HEIGHT attribute.
    *
    * @return   The current value of the HEIGHT attribute.
    */
   public decimal getHeight()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Gets the WIDTH attribute.
    *
    * @return   The current value of the WIDTH attribute.
    */
   public decimal getWidth()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Get the HELP writable attribute.
    * 
    * @return  current value of HELP attribute
    */
   public character getHelp()
   {
      String help = config.getHelp();
      return (help == null) ? new character() : new character(help);
   }

   /**
    * Sets the help text.
    * 
    * @param    help
    *           The new help text.
    */
   public void setHelp(character help)
   {
      setHelp(help.isUnknown() ? "" : help.getValue());
   }

   /**
    * Sets the help text.
    * 
    * @param    help
    *           The new help text.
    */
   public void setHelp(String help)
   {
      config.setHelp(help);
      frame.pushScreenDefinition();
   }

   /**
    * Gets the HIDDEN writable attribute.
    *
    * @return   The current value of the HIDDEN attribute.
    */
   public logical isHidden() 
   {
      return new logical(config.isHidden());
   }

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   public void setHidden(logical hidden) 
   {
      setHidden(hidden.booleanValue());
   }

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   public void setHidden(boolean hidden) 
   {
      config.setHidden(hidden);
      setVisible(!hidden);
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getInnerChars()
   {
      throw new RuntimeException("INNER-CHARS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   ic
    *          Not used.
    */
   public void setInnerChars(integer ic)
   {
      throw new RuntimeException("INNER-CHARS attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getInnerLines()
   {
      throw new RuntimeException("INNER-LINES attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   il
    *          Not used.
    */
   public void setInnerLines(integer il)
   {
      throw new RuntimeException("INNER-LINES attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public logical isKeepFrameZOrder()
   {
      throw new RuntimeException("KEEP-FRAME-Z-ORDER attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   k
    *          Not used.
    */
   public void setKeepFrameZOrder(logical k)
   {
      throw new RuntimeException("KEEP-FRAME-Z-ORDER attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   k
    *          Not used.
    */
   public void setKeepFrameZOrder(boolean k)
   {
      throw new RuntimeException("KEEP-FRAME-Z-ORDER attribute is undefined");
   }

   /**
    * Get the label text for this widget.
    *
    * @return   The label text.
    */
   public character getLabel()
   {
      return new character(config.getLabel());
   }

   /**
    * Get the label text for this widget.
    *
    * @return   The label text.
    */
   public String getLabelStr()
   {
      return config.getColumnLabel() != null ? 
                                              config.getColumnLabel() : 
                                              config.getLabel();
   }
   
   /**
    * Sets the label text, removing any '&' characters first (this is
    * an undocumented feature of Progress with labels).
    * 
    * @param    label
    *           The new label text.
    */
   public void setLabel(character label)
   {
      setLabel((label.isUnknown() ? "" : label.getValue()));
   }

   /**
    * Sets the label text, removing any '&' characters first (this is
    * an undocumented feature of Progress with labels).
    * 
    * @param    label
    *           The new label text.
    */
   public void setLabel(String label)
   {
      // handle case of very long label
      label = StringHelper.removeChar(label, '&');
      
      if (frame.isViewed())
      {
         if (label != null && 
             config.getLabel() != null &&
             label.length() > config.getLabel().length())
         {
            StringBuilder err = new StringBuilder();
            err.append("**Unable to set SCREEN-VALUE.  ")
            .append("LITERAL widget does not fit in FRAME ")
            .append(frame.getName())
            .append(". (4054)");

            LogicalTerminal.message(err.toString());

            return;
         }
      }
      else
      {
         // force setting of frame
         config.setLabel(null);
      }

      config.setLabel(label);
      frame.pushScreenDefinition();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public character getListItems()
   {
      throw new RuntimeException("LIST-ITEMS attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   l
    *          Not used.
    */
   public void setListItems(character l)
   {
      throw new RuntimeException("LIST-ITEMS attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   l
    *          Not used.
    */
   public void setListItems(String l)
   {
      throw new RuntimeException("LIST-ITEMS attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public character getListItemPairs()
   {
      throw new RuntimeException("LIST-ITEM-PAIRS attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.

    * @param   l
    *          Not used.
    */
   public void setListItemPairs(character l)
   {
      throw new RuntimeException("LIST-ITEM-PAIRS attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   l
    *          Not used.
    */
   public void setListItemPairs(String l)
   {
      throw new RuntimeException("LIST-ITEM-PAIRS attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public decimal getMaxHeightChars()
   {
      throw new RuntimeException("MAX-HEIGHT-CHARS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   m
    *          Not used.
    */
   public void setMaxHeightChars(decimal m)
   {
      throw new RuntimeException("MAX-HEIGHT-CHARS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public decimal getMaxWidthChars()
   {
      throw new RuntimeException("MAX-WIDTH-CHARS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   m
    *          Not used.
    */
   public void setMaxWidthChars(decimal m)
   {
      throw new RuntimeException("MAX-WIDTH-CHARS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public logical isMultiple()
   {
      throw new RuntimeException("MULTIPLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   multiple
    *          Not used.
    */
   public void setMultiple(boolean multiple)
   {
      throw new RuntimeException("MULTIPLE attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   multiple
    *          Not used.
    */
   public void setMultiple(logical multiple)
   {
      throw new RuntimeException("MULTIPLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public GenericWidget getNextTabItem()
   {
      throw new RuntimeException("NEXT-TAB-ITEM attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   n
    *          Not used.
    */
   public void setNextTabItem(GenericWidget n)
   {
      throw new RuntimeException("NEXT-TAB-ITEM attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getNumSelectedRows()
   {
      throw new RuntimeException("NUM-SELECTED-ROWS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   r
    *          Not used.
    */
   public void setNumSelectedRows(integer r)
   {
      throw new RuntimeException("NUM-SELECTED-ROWS attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public Object getParent()
   {
      throw new RuntimeException("PARENT attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   p
    *          Not used.
    */
   public void setParent(Object p)
   {
      throw new RuntimeException("PARENT attribute is undefined");
   }

   /**
    * Establishes a relationship between the container and the elements in
    * a given widget group.
    *
    * @param    group
    *           The container of the widget.
    */
   public void setGroup(Group group)
   {
      throw new RuntimeException("Group support not available.");
   }
   
   /**
    * Accesses the first widget in the list of contained widgets.
    *
    * @return   The first widget or <code>null</code> if no widgets are
    *           contained.
    */
   public GenericWidget getFirstChild()
   {
      throw new RuntimeException("FIRST-CHILD attribute is undefined.");
   }

   /**
    * Accesses the last widget in the list of contained widgets.
    *
    * @return   The last widget or <code>null</code> if no widgets are
    *           contained.
    */
   public GenericWidget getLastChild()
   {
      throw new RuntimeException("LAST-CHILD attribute is undefined.");
   }
   
   /**
    * Obtains the next sibling widget in the current field group.
    *
    * @return   Always throws an exception.
    */
   public handle getNextSibling()
   {
      throw new RuntimeException("NEXT-SIBLING attribute is undefined.");
   }
   
   /**
    * Obtains the previous sibling widget in the current field group.
    *
    * @return   Always throws an exception.
    */
   public handle getPrevSibling()
   {
      throw new RuntimeException("PREV-SIBLING attribute is undefined.");
   }
   
   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getPfColor()
   {
      throw new RuntimeException("PFCOLOR attribute is undefined");
   }

   /**
    * Sets the input (prompt/edit mode) color.
    *
    * @param    param
    *           The color to use.
    */
   public void setPfColor(NumberType param)
   {
      config.setPfcolor(new ColorSpec(param.intValue()).convert());
      frame.pushScreenDefinition();
   }

   /**
    * Sets the input (prompt/edit mode) color.
    * 
    * @param    inputColor
    *           The color to use for when in "input" mode.
    */
   public void setPfcolor(Color inputColor)
   {
      config.setPfcolor(inputColor);
      frame.pushScreenDefinition();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public character getPrivateData()
   {
      throw new RuntimeException("PRIVATE-DATA attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   d
    *          Not used.
    */
   public void setPrivateData(character d)
   {
      throw new RuntimeException("PRIVATE-DATA attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   d
    *          Not used.
    */
   public void setPrivateData(String d)
   {
      throw new RuntimeException("PRIVATE-DATA attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public P2JQuery getQuery()
   {
      throw new RuntimeException("QUERY attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   q
    *          Not used.
    */
   public void setQuery(P2JQuery q)
   {
      throw new RuntimeException("QUERY attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public decimal getRow()
   {
      throw new RuntimeException("ROW attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   row
    *          Not used.
    */
   public void setRow(NumberType row)
   {
      throw new RuntimeException("ROW attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   row
    *          Not used.
    */
   public void setRow(double row)
   {
      throw new RuntimeException("ROW attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public logical isScrollable()
   {
      throw new RuntimeException("SCROLLABLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   public void setScrollable(logical s)
   {
      throw new RuntimeException("SCROLLABLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   public void setScrollable(boolean s)
   {
      throw new RuntimeException("SCROLLABLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public logical isSelected()
   {
      throw new RuntimeException("SELECTED attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   public void setSelected(logical s)
   {
      throw new RuntimeException("SELECTED attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   s
    *          Not used.
    */
   public void setSelected(boolean s)
   {
      throw new RuntimeException("SELECTED attribute is undefined");
   }

   /**
    * Queries the SENSITIVE attribute of this widget.
    *
    * @return <code>true</code> if SENSITIVE is set.
    */
   public logical isSensitive()
   {
      boolean sen = LogicalTerminal.isSensitive(this);
      return new logical(sen);
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   public void setSensitive(logical s)
   {
      setSensitive(s.booleanValue());
   }
   
   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   public void setSensitive(String s)
   {
      setSensitive(new logical(s).booleanValue());
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   public void setSensitive(character s)
   {
      setSensitive(new logical(s).booleanValue());
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   public void setSensitive(boolean s)
   {
      LogicalTerminal.setSensitive(this, s);
   }

   /**
    * Queries the DATA-TYPE attribute of this widget.
    *
    * @return A character value that represents the data type of the field
    *         associated with the widget.
    */
   public character getDataType()
   {
      return new character(config.getDataType());
   }

   /**
    * Sets new value of DATA-TYPE attribute.
    *
    * @param dataType
    *        new value of DATA-TYPE attribute
    */
   public void setDataType(character dataType)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Queries the TABLE attribute of this widget.
    *
    * @return The name of the database table containing the field associated
    *         with the widget.
    */
   public character getTable()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Queries the TOOLTIP attribute of this widget.
    *
    * @return A help text message for a text field or text variable.
    */
   public character getTooltip()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets new value of TOOLTIP attribute.
    *
    * @param tooltip
    *        new value of TOOLTIP attribute
    */
   public void setTooltip(character tooltip)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets new value of TOOLTIP attribute.
    *
    * @param tooltip
    *        new value of TOOLTIP attribute
    */
   public void setTooltip(String tooltip)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Queries the WIDTH-PIXELS attribute of this widget.
    *
    * @return  The width of the widget in pixels.
    */
   public integer getWidthPixels()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets new value of WIDTH-PIXELS attribute.
    *
    * @param widthPixels
    *        new value of WIDTH-PIXELS attribute
    */
   public void setWidthPixels(NumberType widthPixels)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Queries the HEIGHT-PIXELS attribute of this widget.
    *
    * @return  The height of the widget in pixels.
    */
   public integer getHeightPixels()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param widthPixels
    *        new value of HEIGHT-PIXELS attribute
    */
   public void setHeightPixels(NumberType widthPixels)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets the TAB-STOP option.
    * 
    * @param    tabStop
    *           <code>true</code> if the option is set to ON.
    */
   public void setTabStop(boolean tabStop)
   {
      // valid for descendants of BaseConfig only
      ((BaseConfig) config).setTabStop(tabStop);
      frame.pushScreenDefinition();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public character getTitle() 
   {
      throw new RuntimeException("TITLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   t
    *          Not used.
    */
   public void setTitle(character t) 
   {
      throw new RuntimeException("TITLE attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   t
    *          Not used.
    */
   public void setTitle(String t)
   {
      throw new RuntimeException("TITLE attribute is undefined");
   }
   
   /**
    * Obtains the type of widget description.
    *
    * @return   Always returns <code>unknown value</code>.
    */
   public character getType()
   {
      return new character();
   }

   /**
    * Registers a validation method for any input into this widget.
    * 
    * @param    validator
    *           An instance of <code>Validateable</code> interface to call to
    *           perform input validation.
    */
   public void setValidatable(Validatable validator)
   {
      this.validator = validator;
   }

   /**
    * Checks if the widget is validateable.
    * 
    * @return   <code>true</code> if the widget is validateable.
    */
   public boolean isValidatable()
   {
      return validator != null;
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   direction
    *          Not used.
    */
   public void setVertical(boolean direction)
   {
      throw new RuntimeException("setVertical attribute is undefined");
   }

   /**
    * Get the VISIBLE writable attribute.
    *
    * @return   The current value of VISIBLE attribute.
    */
   public logical isVisible()
   {
      return new logical(config.isVisible());
   }

   /**
    * Set the VISIBLE writable attribute.
    *
    * @param    visible
    *           The new value for the VISIBLE attribute.
    */
   public void setVisible(logical visible)
   {
      setVisible(visible.booleanValue());
   }

   /**
    * Set the VISIBLE writable attribute.
    *
    * @param    visible
    *           The new value for the VISIBLE attribute.
    */
   public void setVisible(boolean visible)
   {
      config.setVisible(visible);
      config.setHidden(!visible);
      frame.pushScreenDefinition();
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getX()
   {
      throw new RuntimeException("X attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   x
    *          Not used.
    */
   public void setX(integer x)
   {
      throw new RuntimeException("X attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @return  nothing.
    */
   public integer getY()
   {
      throw new RuntimeException("Y attribute is undefined");
   }

   /**
    * Blocks attempts to access this attribute.
    * 
    * @param   y
    *          Not used.
    */
   public void setY(integer y)
   {
      throw new RuntimeException("Y attribute is undefined");
   }
   
   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   itemList
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical addFirst(character itemList)
   {
      throw new RuntimeException("ADD-FIRST() method is undefined");
   }
   
   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   itemList
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical addFirst(String itemList)
   {
      throw new RuntimeException("ADD-FIRST() method is undefined");
   }
   
   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   label
    *          Not used.
    * @param   value
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical addFirst(character label, character value)
   {
      throw new RuntimeException("ADD-FIRST() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   label
    *          Not used.
    * @param   value
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical addFirst(String label, String value)
   {
      throw new RuntimeException("ADD-FIRST() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   index
    *          Not used.
    * 
    * @return  nothing.
    */
   public character entry(integer index)
   {
      throw new RuntimeException("ENTRY() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   item
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical isSelected(character item)
   {
      throw new RuntimeException("IS-SELECTED() method is undefined");
   }
   
   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   index
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical isSelected(integer index)
   {
      throw new RuntimeException("IS-SELECTED() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   text
    *          Not used.
    * 
    * @return  nothing.
    */
   public integer lookup(character text)
   {
      throw new RuntimeException("LOOKUP() method is undefined");
   }
   
   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   text
    *          Not used.
    * 
    * @return  nothing.
    */
   public integer lookup(String text)
   {
      throw new RuntimeException("LOOKUP() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public integer getNumColumns()
   {
      throw new RuntimeException("NUM-COLUMNS attribute is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   index
    *          Not used.
    * 
    * @return  nothing.
    */
   public BrowseColumnWidget getBrowseColumn(double index)
   {
      throw new RuntimeException("GET-BROWSE-COLUMN method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   index
    *          Not used.
    * 
    * @return  nothing.
    */
   public BrowseColumnWidget getBrowseColumn(NumberType index)
   {
      throw new RuntimeException("GET-BROWSE-COLUMN method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public logical selectAll()
   {
      throw new RuntimeException("SELECT-ALL() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public logical deselectRows()
   {
      throw new RuntimeException("DESELECT-ROWS() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   n
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical fetchSelectedRow(integer n)
   {
      throw new RuntimeException("FETCH-SELECTED-ROW() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public logical refresh()
   {
      throw new RuntimeException("REFRESH() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @return  nothing.
    */
   public logical selectFocusedRow()
   {
      throw new RuntimeException("SELECT-FOCUSED-ROW() method is undefined");
   }

   /**
    * Blocks attempts to access this widget method.
    * 
    * @param   n
    *          Not used.
    * @param   conditional
    *          Not used.
    * 
    * @return  nothing.
    */
   public logical setRepositionedRow(integer n, boolean conditional)
   {
      throw new RuntimeException("SET-REPOSITIONED-ROW() method is " +
                                 "undefined");
   }

   /**
    * Gets the widget name.
    * 
    * @return   The widget name or unknown value if not configured.
    */
   public character getName()
   {
      if (config == null)
         return new character();
      
      return new character(frame.getName(this));
   }

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    event
    *           The name of the event.
    */
   public void apply(character event)
   {
      if (event.isUnknown())
         return;
      
      apply(event.toStringMessage());
   }

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    event
    *           The name of the event.
    */
   public void apply(String event)
   {
      LogicalTerminal.apply(frame, getId(), event);
   }

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    eventCode
    *           The event code.
    */
   public void apply(integer eventCode)
   {
      if (eventCode.isUnknown())
         return;
      
      apply(eventCode.intValue());
   }

   /**
    * Applies an event to this widget.
    * <p>
    * This method corresponds to the Progress statement: APPLY widget
    * 
    * @param    eventCode
    *           The event code.
    */
   public void apply(double eventCode)
   {
      LogicalTerminal.apply(frame, getId(), (int)eventCode);
   }

   /**
    * COLOR DISPLAY
    * 
    * @param   cs
    *          Color specification.
    */
   public void colorDisplay(ColorSpec cs)
   {
      setDcolor(cs.convert());
   }

   /**
    * COLOR PROMPT
    * 
    * @param   cs
    *          Color specification.
    */
   public void colorPrompt(ColorSpec cs)
   {
      setPfcolor(cs.convert());
   }

   /**
    * Save configured colors.
    */
   public void saveColors()
   {
      config.saveColors();
   }
   
   /**
    * Restore configured colors.
    */
   public void restoreColors()
   {
      config.restoreColors();
   }
   
   /**
    * Disables this widget.
    * <p>
    * This method corresponds to the Progress statement: DISABLE widget
    */
   public void disable()
   {
      LogicalTerminal.disable(getFrame(), new int[] {getId()});
   }

   /**
    * Enables this widget.
    * <p>
    * This method corresponds to the Progress statement: ENABLE widget
    */
   public void enable()                  
   {
      config.setVisible(true);
      LogicalTerminal.enable(getFrame(), new int[] {getId()});
   }

   /**
    * Hides this widget.
    * <p>
    * This method corresponds to the Progress statement: HIDE widget
    */
   public void hide()
   {
      hide(false);
   }

   /**
    * Hides this widget.
    * <p>
    * This method corresponds to the Progress statement: HIDE widget
    * 
    * @param   noPause
    *          <code>true</code> causes omission of any pauses which would
    *          be necessary otherwise.
    */
   public void hide(boolean noPause)
   {
      config.setVisible(false); 
      LogicalTerminal.hide(getId(), noPause);
   }

   /**
    * Underlines this widget on the screen.
    * <p>
    * This method corresponds to the Progress statement: UNDERLINE
    */
   public void underline()
   {
      getFrame().underline(new int[] {getId()});
   }

   /**
    * Underlines this widget on the screen.
    * <p>
    * This method corresponds to the Progress statement: UNDERLINE
    *
    * @param    stream
    *           An explicit stream which will be used for the output.
    */
   public void underline(Stream stream)
   {
      getFrame().underline(stream, new int[] {getId()});
   }

   /**
    * Shows this widget.
    * <p>
    * This method corresponds to the Progress statement: VIEW widget
    */
   public void view()
   {
      config.setVisible(true); 
      LogicalTerminal.view(getId());
   }
   
   /**
    * Reports if this widget object is valid for use.
    * <p>
    * The validity of the widget object is determined solely by the scope of
    * the containing frame.
    * 
    * @return   <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean isValid()
   {
      if (frame == null)
         return false;
      
      return frame.isValid();
   }

   /**
    * Reports if this object is unknown.<p>
    * This implementation always returns <code>false</code>.
    *
    * @return   <code>false</code>.
    */
   @Override
   public boolean isUnknown()
   {
      return false;
   }

   /**
    * Processes validation rules (if they exist), error checking and any
    * other constraint processing that needs to be applied.  This is used
    * to check if a proposed update to an object will pass validation.
    *
    * @param    val
    *           The new value that is proposed for assignment to the object.
    *
    * @return   <code>null</code> if validation succeeds, error message
    *           otherwise.
    */
   public String validate(BaseDataType val)
   {
      if (validator == null)
         return "No validation method defined";

      try
      {
         validator.validate(val);
      }
      catch (ErrorConditionException ece)
      {
         return ece.getMessage();
      }
      catch (UnsupportedOperationException uoe)
      {
         return "This widget is non-editable";
      }

      return null;
   }

   /**
    * Gets the configuration object that defines the appearance and
    * functioning of this widget.
    * 
    * @return   <code>ComponentConfig</code> instance.
    */
   ComponentConfig getConfig()
   {
      return config;
   }

   /**
    * Sets the format specification.
    * 
    * @param    format
    *           The format specification.
    * @param    doPush
    *           Force pushing screen definition.
    */
   void setFormat(String format, boolean doPush)
   {
      config.setFormat(format);
      
      if (!frame.isBatch())
         LogicalTerminal.setFormat(config.getID(), format);
   }

   /**
    * Mark widget as a header.
    * 
    * @param    header
    *           New value of the flag.
    */
   public void setHeader(boolean header)
   {
      config.setHeader(header);
   }
   
   /**
    * Obtain the current value in the screen buffer of the backing data for
    * this widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   public character getScreenValue()
   {
      return frame.getScreenValue(this);
   }
   
   /**
    * Set the current value in the screen buffer of the backing data for
    * this widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   public void setScreenValue(character value)   
   {
      frame.setScreenValue(this, value);
   }

   /**
    * This method implements MOVE-TO-TOP function.
    *
    * @return  <code>true</code> if operation was successful.
    */
   public logical moveToTop()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * This method implements MOVE-TO-BOTTOM function.
    *
    * @return  <code>true</code> if operation was successful.
    */
   public logical moveToBottom()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * This method implements MOVE-AFTER-TAB function.
    *
    * @param   widget
    *          Widget reference.
    *
    * @return  <code>true</code> if operation was successful.
    */
   public logical moveAfterTab(GenericWidget widget)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Get this widget wrapped in a handle.
    * 
    * @return  handle to this frame.
    */
   public handle asWidgetHandle()
   {
      return new handle(this);
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    */
   public void readOnlyError(String attribute)
   {
      handle.readOnlyError(asWidgetHandle(), attribute);
   }
}
